import twitter

from zope.interface import implements

from interfaces import ITwitterAPI

def wrapper(func_name, username, password, *args, **kwargs):
    api = twitter.Api(username, password)
    func = getattr(api, func_name)
    return func(*args, **kwargs)
 
class TwitterAPI(object):
    """
    Abstract python-twitter API. This class effectively modifies
    every callable method signature in python-twitter to expect
    username and password as parameters.
    """
    implements(ITwitterAPI)
  
    # List to lazily build a set of callable functions in
    # python-twitter
    __callable_functions__ = []
    
    def __getattr__(self, name):
        if hasattr(twitter.Api, name):           
            attr = getattr(twitter.Api, name)
            if callable(attr):
                if name not in self.__callable_functions__:
                    self.__callable_functions__.append(name)
                return lambda username, password, *args, **kwargs: wrapper(name, username, password, *args, **kwargs)
            else:
                return attr
        return getattr(object, name)



from zope.interface import implements
from interfaces import ITwitterCredentialsProvider

class MemberDataTwitterCredentialsProvider:

    implements(ITwitterCredentialsProvider)

    def __init__(self, context):
        self.context = context

    def username(self):
        return self.context.getProperty('twitterUsername')

    def password(self):
        return self.context.getProperty('twitterPassword')

class ReMemberTwitterCredentialsProvider:

    implements(ITwitterCredentialsProvider)

    def __init__(self, context):
        self.context = context

    # Do not use ClassGen accessors since they do not work if the fields
    # are provided by schema extension
    def username(self):
        return self.context.Schema().getField('twitterUsername').getAccessor(self.context)()

    def password(self):
        return self.context.Schema().getField('twitterPassword').getAccessor(self.context)()




import types
from urllib2 import HTTPError, URLError
from twitter import TwitterError

from Acquisition import aq_inner, aq_parent
from Acquisition import Implicit
from zope.interface import implements
from zope.component import getAdapters, queryMultiAdapter, getUtility
from zope.viewlet.interfaces import IViewletManager, IViewlet
from kss.core import kssaction, force_unicode
from plone.app.kss.interfaces import IPloneKSSView
from plone.app.kss.plonekssview import PloneKSSView

from Products.CMFCore.utils import getToolByName
from Products.CMFPlone import PloneMessageFactory as _

from interfaces import ITwitterAPI

# We cannot use a normal decorator since there are issues with marshalling.
# Luckily kss.core provides the kssaction decorator which we can subclass
# in a creative way. It is easier than attempting to chain decorators.
class kssactionplus(kssaction):

    def apply(self, obj, *arg, **kw):
        try:
            # Clear our own error messages
            ksscore = obj.getCommandSet('core')
            selector = ksscore.getHtmlIdSelector('collective-sylvester-dashboard-messages')
            ksscore.replaceInnerHTML(selector, '')
            
            # Clear Plone's messages
            kssplone = obj.getCommandSet('plone')
            kssplone.issuePortalMessage('')

            return kssaction.apply(self, obj, *arg, **kw)

        except TwitterError:
            viewname = '@@collective.sylvester.service-error'
        except URLError:
            viewname = '@@collective.sylvester.service-error'
        except HTTPError, e:
            if e.msg == 'Unauthorized':
                viewname = '@@collective.sylvester.auth-error'
            else:
                viewname = '@@collective.sylvester.unknown-error'
        except:
            viewname = '@@collective.sylvester.unknown-error'
       
        html = obj.macroContent('%s/macros/main' % viewname)        
        kssplone = obj.getCommandSet('plone')
        kssplone.issuePortalMessage(html, msgtype='error')
        return obj.render()

class SylvesterKssView(Implicit, PloneKSSView):
    """
    Methods _macroContent, viewletMacroContent, macrocontent from 
    https://svn.plone.org/svn/plone/plone.app.kss/branches/hedley-macrocontent
    included. I omit the docstrings.
    """

    implements(IPloneKSSView)

    def _macroContent(self, provider, macro_name, context=None, **kw):
        # Determine context to use for rendering
        if context is None:
            render_context = aq_inner(self.context)
        else:
            render_context = context

        # Build extra context. These variables will be in
        # scope for the macro.        
        extra_context = {'options':{}}
        extra_context.update(kw)
        the_macro = None

        # Determine what type of provider we are dealing with
        if isinstance(provider, types.StringType):
            # Page template or browser view. Traversal required.
            pt_or_view = render_context.restrictedTraverse(provider)
            if provider.startswith('@@'):            
                the_macro = pt_or_view.index.macros[macro_name]
            else:          
                the_macro = pt_or_view.macros[macro_name]

            # template_id seems to be needed, so add to options
            # if it is not there
            if not extra_context['options'].has_key('template_id'):
                extra_context['options']['template_id'] = provider.split('/')[-1]

        elif IViewlet.isImplementedBy(provider) or IPortletRenderer.isImplementedBy(provider):
            the_macro = provider.render.macros[macro_name]
        
        # Adhere to header_macros convention. Setting the_macro here
        # ensures that code calling this method cannot override the_macro.
        extra_context['options']['the_macro'] = the_macro

        # If context is explicitly passed in then make available        
        if context is not None:
            extra_context['context'] = context

        if hasattr(self, 'header_macros'):
            # plone.app.kss <= 1.4.3
            content = self.header_macros.__of__(render_context).pt_render(
                        extra_context=extra_context)
        else:
            # plone.app.kss > 1.4.3
            from plone.app.kss.plonekssview import header_macros
            content = header_macros.__of__(self).__of__(render_context).pt_render(
                        extra_context=extra_context)

        # IE6 has problems with whitespace at the beginning of content
        content = content.strip()

        # Always encoded as utf-8
        content = force_unicode(content, 'utf')
        return content

    def viewletMacroContent(self, manager_name, viewlet_name, macro_name, 
        context=None, **kw):
        manager = queryMultiAdapter(
            (self.context, self.request, self), 
            IViewletManager, 
            name=manager_name)

        viewlets = getAdapters(
            (manager.context, manager.request, manager.__parent__, manager), 
            IViewlet)

        target = None
        for order, (name, viewlet) in enumerate(viewlets):
            if name == viewlet_name:               
                return self._macroContent(viewlet, macro_name, context, **kw)

        # Raise on lookup error
        msg = "No viewlet %s registered with manager %s" % \
            (viewlet_name, manager_name)
        raise RuntimeError, msg

    def macroContent(self, macropath, **kw):
        'Renders a macro and returns its text'
        path = macropath.split('/')
        if len(path) < 2 or path[-2] != 'macros':
            raise RuntimeError, 'Path must end with macros/name_of_macro (%s)' % (repr(macropath), )
        # needs string, do not tolerate unicode (causes but at traverse)
        jointpath = '/'.join(path[:-2]).encode('ascii')

        # put parameters on the request, by saving the original context
        self.request.form, orig_form = kw, self.request.form
        content = self._macroContent(
                    provider=jointpath, 
                    macro_name=path[-1],                  
                    **kw
                    )
        self.request.form = orig_form

        return content

    @kssactionplus
    def tweet(self): 
        context = aq_inner(self.context)
        view = context.restrictedTraverse('@@collective.sylvester.dashboard')

        message = context.REQUEST.form['message'].strip()
        view.twitter.PostUpdate(message, handle_error=False)

        # xxx: caching issue! python-twitter caches results for a minute
        html = self.macroContent(
                '@@collective.sylvester.dashboard/macros/update_form',
                view=view)

        ksscore = self.getCommandSet('core')
        selector = ksscore.getHtmlIdSelector('collective-sylvester-dashboard-tweet-form')
        ksscore.replaceHTML(selector, html)

        kssplone = self.getCommandSet('plone')
        msg = _("Your message has been sent")
        kssplone.issuePortalMessage(msg, msgtype='info')

    @kssactionplus
    def dashboardNavigate(self, item): 
        """
        Load content area of dashboard.

        Param item is the name of a macro in the dashboard view template.
        """
        context = aq_inner(self.context)        
        view = context.restrictedTraverse('@@collective.sylvester.dashboard')
        html = self.macroContent(
            '@@collective.sylvester.dashboard/macros/%s' % item,
            view=view)

        ksscore = self.getCommandSet('core')
        selector = ksscore.getHtmlIdSelector(
            'collective-sylvester-dashboard-content')
        ksscore.replaceInnerHTML(selector, html)

        # Update the menus
        selector = ksscore.getCssSelector(
            'div.collective-sylvester-dashboard-navigation a')
        ksscore.removeClass(selector, 'selected')
        selector = ksscore.getHtmlIdSelector(
            'collective-sylvester-dashboard-navigation-%s' % item)
        ksscore.addClass(selector, 'selected')

    @kssactionplus
    def refreshFriendlet(self, friend_id): 
        """
        Refresh a single friendlet

        xxx: not currently used
        """
        context = aq_inner(self.context)
        ksscore = self.getCommandSet('core')

        view = context.restrictedTraverse('@@collective.sylvester.dashboard')
        friend = view.twitter.GetUser(friend_id, handle_error=False)
        html = self.viewletMacroContent(                    
                'collective.sylvester.friendletmanager', 
                'collective.sylvester.friendlet', 
                'body',
                friend=friend)
        selector = ksscore.getHtmlIdSelector(
            'collective-sylvester-friendlet-friend%s' % friend_id) 
        ksscore.replaceHTML(selector, html)

    @kssactionplus
    def refreshFriendlets(self, friend_ids=None): 
        """
        Refresh a a few friendlets.
        """
        context = aq_inner(self.context)
        ksscore = self.getCommandSet('core')

        view = context.restrictedTraverse('@@collective.sylvester.dashboard')
        mapping = view.GetFriendsStatuses(user_ids=friend_ids)
        for di in mapping.items():
            user = di['user']
            statuses = ['statuses']
            html = self.viewletMacroContent(                    
                    'collective.sylvester.friendletmanager', 
                    'collective.sylvester.friendlet', 
                    'main',
                    friend=user,
                    statuses=statuses)
            selector = ksscore.getHtmlIdSelector(
                'collective-sylvester-friendlet-friend%s' % friend_id) 
            ksscore.replaceHTML(selector, html)

    @kssaction
    def cachewarmer(self):
        """
        Do a few common calls to twitter. The results are not important.
        Doing these calls periodically ensures that results are always 
        availablable from RAM.
        """
        context = aq_inner(self.context)
        view = context.restrictedTraverse('@@collective.sylvester.dashboard')
        # These methods are RAM-cached by the view
        view.GetFriendsTimeline()
        view.GetReplies()



from time import time
import re
from urllib import urlencode
from urllib2 import urlopen, HTTPError, URLError
from twitter import TwitterError
try:
    from stripogram import html2text
    HAS_STRIPOGRAM = True
except ImportError:
    HAS_STRIPOGRAM = False
from DateTime import DateTime
from zope.interface import implements
from zope.component import getUtility, getAdapter
from Acquisition import aq_inner
from Products.PythonScripts.standard import url_quote
from Products.Five.browser import BrowserView
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone import PloneMessageFactory as _
from plone import memoize

from interfaces import ISylvesterView, ICredentialsFormView, \
    IPublishToTwitterFormView, ITwitterAPI, ITwitterCredentialsProvider

TEXT_URL_TO_ANCHOR = re.compile(r'(((f|ht){1}tp://)[-a-zA-Z0-9@:%_\+.~#?&//=]+)', re.I|re.M)

# Handy decorator to catch exceptions and present useful
# error messages
def errorhandler(func):

    def new(self, *args, **kwargs):        
        # If handle_error is False then this decorator must let exceptions
        # propagate.
        handle_error = kwargs.get('handle_error', True)
        if not handle_error:
            return func(self, *args)

        viewname = ''
        try:
            return func(self, *args)
        except TwitterError:
            viewname = '@@collective.sylvester.service-error'
        except URLError:
            viewname = '@@collective.sylvester.service-error'
        except HTTPError, e:
            if e.msg == 'Unauthorized':
                viewname = '@@collective.sylvester.auth-error'
            else:
                viewname = '@@collective.sylvester.unknown-error'
        except:
            viewname = '@@collective.sylvester.unknown-error'

        context = aq_inner(self.context)
        self.request.RESPONSE.redirect('%s/%s' % (context.absolute_url(), viewname))

    return new

class TwitterAPIAdapter(object):
    """
    Utilities are unique in-memory objects which means we
    cannot store authentication information on them. We 
    need this adapter to provide authentication information 
    to the TwitterAPI utility.
    """

    def __init__(self, context, username, password):
        self.context = context
        self.username = username
        self.password = password

    def wrapper(self, func_name, *args, **kwargs):
        #import logging
        #logger = logging.getLogger('sylvester')
        #logger.info("doing a call to twitter %s" % func_name)
        func = getattr(self.context, func_name)
        # handle_error must be removed from kwargs since it is not known
        # to python-twitter.
        kw = kwargs.copy()
        if kw.has_key('handle_error'):
            del kw['handle_error']
        return func(self.username, self.password, *args, **kw)

    def __getattr__(self, name):           
        if hasattr(self.context, name):
            if name in self.context.__callable_functions__:
                return lambda *args, **kwargs: self.wrapper(name, *args, **kwargs)
            else:
                return getattr(self.context, name)
        return getattr(object, name)

class SylvesterView(BrowserView):

    implements(ISylvesterView)

    twitter = None

    def __init__(self, context, request):
        BrowserView.__init__(self, context, request)
        self._setup_twitter_adapter()

    def name(self):
        return self.__name__

    def authenticate(self, redirect=False):
        res = self._setup_twitter_adapter()
        if not res and redirect:
            self.redirect_to_credentials_form()
        return res

    def redirect_to_credentials_form(self):
        #Redirect to page where user can enter credentials for this session.
        portal = getToolByName(self, 'portal_url').getPortalObject()
        self.request.response.redirect(
                '%s/@@collective.sylvester.credentials-form?came_from=%s' \
                % (portal.absolute_url(), url_quote(self.request.URL)))

    def _setup_twitter_adapter(self):      
        username, password = 'unlikely', 'u;n;l;i;k;e;l;y'
        # Check if the session is storing credentials
        session = self.request.SESSION
        if session.has_key('collective.sylvester.twitter'):
            username = session['collective.sylvester.twitter']['username']
            password = session['collective.sylvester.twitter']['password']                
        else:
            # Find an adapter which can retrieve twitter credentials for the 
            # authenticated member.
            pms = getToolByName(self.context, 'portal_membership')
            if not pms.isAnonymousUser(): 
                member = pms.getAuthenticatedMember()
                adapter = getAdapter(member, ITwitterCredentialsProvider)
                if adapter is not None:            
                    username, password = adapter.username(), adapter.password()

        # Return false if username lookup failed
        if (username == 'unlikely') and (password == 'u;n;l;i;k;e;l;y'):
            return False

        # Setup twitter attribute
        self.twitter = TwitterAPIAdapter(
                            getUtility(ITwitterAPI),
                            username, password)

        # Test authentication. GetFriends is the lightest method I could find
        # for this use-case and this class will cache the result on the request.
        try:
            self.GetFriends(user=None)
        except HTTPError, e:
            # Catch only unauthorized
            if e.msg == 'Unauthorized':
                return False            
            raise

        return True

    def linkify(self, text):
        return TEXT_URL_TO_ANCHOR.sub('<a href="\g<1>">\g<1></a>', text)
    
    def format_ago(self, atime, uppercase=False, invert=False):                
        try:
            dt = DateTime(atime)
        except ValueError:
            return None

        now = DateTime().toZone('GMT+0')
        now_secs = float(now)
        dt_secs = float(dt) 

        if not invert:
            # This case is for 'X time ago'. If smaller than zero just set to 
            # zero.
            diff = now_secs - dt_secs    
            if diff < 0:
                diff = 0
        else:
            # This case is for 'X time left'. If smaller than zero then no time 
            # left, so return None.
            diff = dt_secs - now_secs
            if diff <= 0:
                return None

        ret = ''
        if diff < 3600:
            minutes = int(diff / 60)
            if minutes == 1:
                ret = _('1 minute')
            else:
                ret = _('%s minutes' % minutes)
        elif (diff >= 3600) and (diff < 86400):
            hours = int(diff / 3600)
            if hours == 1:
                ret = _('1 hour')
            else:
                ret = _('%s hours' % hours)
        elif diff >= 86400:
            days = int(diff / 86400)
            if days == 1:
                ret = _('1 day')
            else:
                ret = _('%s days' % days)

        if uppercase:
            return ret.upper()

        return ret

    def latest(self):
        result = self.twitter.GetUserTimeline(count=1)
        if result:
            return result[0]
        return None

    @memoize.ram.cache(lambda *args: time() // 60)
    def GetFriendsTimeline(self, user=None, since=None):
        """
        This call is potentially too slow for normal use in ajax requests.
        The dashboard makes this call when it is first loaded, which provides
        an opportunity to use a memoize decorator.
        """
        return self.twitter.GetFriendsTimeline(user=user, since=since)

    @memoize.ram.cache(lambda *args: time() // 60)
    def GetReplies(self):
        """
        Cached for the same reason as GetFriendsTimeline
        """
        return self.twitter.GetReplies()

    def GetFriendsStatuses(self, user_ids=None):
        """
        Take the result from GetFriendsTimeline and group the statuses by user.
        Return a list of dictionaries of the form 
        {'user':user, 'statuses':statuses}
        """
        user_status_map = {}
        userid_user_map = {}
        for status in self.GetFriendsTimeline():
            # Skip over users we are not interested in
            if (user_ids is not None) and (status.user.id not in user_ids):
                continue
            user_status_map.setdefault(status.user.id, [])
            user_status_map[status.user.id].append(status)
            userid_user_map[status.user.id] = status.user
        
        result = []
        for userid, statuses in user_status_map.items():
            result.append(dict(user=userid_user_map[userid], statuses=statuses))

        # Sort
        def mysort(a, b):
            return cmp(a['user'].screen_name, b['user'].screen_name)
        result.sort(mysort)

        return result

    @memoize.request.cache(
        lambda func, obj, user: (func.func_name, user), 
        get_request='self.request')
    def GetFriends(self, user=None):
        return self.twitter.GetFriends(user=user)

    @errorhandler
    def __call__(self, *args, **kwargs):
        if not self.authenticate(redirect=True):
            return 

        # The template is cleaner if we disable the border here
        self.request.set('disable_border', 1)

        return super(SylvesterView, self).__call__(*args, **kwargs)

class CredentialsFormView(BrowserView):

    implements(ICredentialsFormView)

    def name(self):
        return self.__name__

    def submit(self, username, password, came_from=''):
        session = self.request.SESSION
        if not session.has_key('collective.sylvester.twitter'):
            session['collective.sylvester.twitter'] = {}
        session['collective.sylvester.twitter']['username'] = username
        session['collective.sylvester.twitter']['password'] = password

        getToolByName(self, 'plone_utils').addPortalMessage(
            _("Twitter credentials saved for this session"))
        if came_from:
            self.request.response.redirect(came_from)
        else:
            portal = getToolByName(self, 'portal_url').getPortalObject()
            self.request.response.redirect(portal.absolute_url())

    def __call__(self, *args, **kwargs):
        form = self.request.form
        if form.has_key('submit'):
            return self.submit(
                form.get('username', ''), 
                form.get('password', ''),
                form.get('came_from', ''))

        self.request.set('disable_border', 1)
        return super(CredentialsFormView, self).__call__(*args, **kwargs)

class PublishToTwitterFormView(BrowserView):

    implements(IPublishToTwitterFormView)

    def name(self):
        return self.__name__

    @errorhandler
    def submit(self, came_from=''):
        # Build up a string from the Description field if it has a value.If 
        # it is empty then use a stripped Text field if present. If both yield 
        # an empty string then use only a tinyurl link.
        context = aq_inner(self.context)
        text = context.Description()
        if not text and HAS_STRIPOGRAM:
            if context.Schema().getField('text') is not None:
                text = html2text(context.getText())
                
        # Get a tiny url link
        url_data = urlencode(dict(url=context.absolute_url()))
        try:
            link = urlopen(
                'http://tinyurl.com/api-create.php', 
                data=url_data).read().strip()
        except URLError:
            msg = "An error was encountered while attempting to shorten the URL"
            getToolByName(self, 'plone_utils').addPortalMessage(_(msg), 'error')
            self.request.response.redirect(context.absolute_url())
            return

        # Reduce length of text
        text = text[:(137-len(link))]

        # Send to twitter
        view = context.restrictedTraverse('@@collective.sylvester.dashboard')
        message = '%s...%s' % (text, link)
        view.twitter.PostUpdate(message, handle_error=False)
        getToolByName(self, 'plone_utils').addPortalMessage(
            _("Item published to twitter"))
        self.request.response.redirect(context.absolute_url())

    def __call__(self, *args, **kwargs):
        form = self.request.form
        if form.has_key('submit'):
            return self.submit(
                form.get('came_from', ''))

        # Redirect if member cannot authenticate to twitter
        context = aq_inner(self.context)
        view = context.restrictedTraverse('@@collective.sylvester.dashboard')
        if not view.authenticate(redirect=True):
            return

        self.request.set('disable_border', 1)
        return super(PublishToTwitterFormView, self).__call__(*args, **kwargs)



from zope.interface import implements
from zope.viewlet.interfaces import IViewlet
from Products.Five.browser import BrowserView
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile

class Friendlet(BrowserView):

    implements(IViewlet)
  
    def __init__(self, context, request, view, manager):
        self.context = context
        self.request = request
        self.__parent__ = view
        self.manager = manager

    def update(self):
        pass

    render = ViewPageTemplateFile("templates/friendlet.pt")    

class Statuslet(BrowserView):

    implements(IViewlet)
  
    def __init__(self, context, request, view, manager):
        self.context = context
        self.request = request
        self.__parent__ = view
        self.manager = manager

    def update(self):
        pass

    render = ViewPageTemplateFile("templates/statuslet.pt")    



from zope.interface import Interface
from zope.viewlet.interfaces import IViewletManager

class ITwitterAPI(Interface):
    """
    Interface for python-twitter API
    """

class ITwitterCredentialsProvider(Interface):
    """
    Interface for adapter which provides twitter credentials
    """
   
    def username():
        """
        Return twitter username
        """

    def password():
        """
        Return twitter password
        """

class ISylvesterView(Interface):
    """
    Interface
    """

    def authenticate(redirect=False):
        """
        Call redirect_to_credentials_form if _setup_twitter_adapter returns 
        False.
        """

    def redirect_to_credentials_form():
        """
        Redirect to page where user can enter credentials for this session.
        """

    def _setup_twitter_adapter():      
        """
        Set up attribute on the view which carries authentication information.
        Return True if member can authenticate to twitter, False otherwise
        """

    def linkify(text):
        """
        Replace occurences of URI's in text with proper markup. Return the
        transformed text.
        """

    def format_ago(atime, uppercase=False, invert=False):
        """        
        Return a relative translated time string, eg. "4 hours ago" or
        "10 minutes ago".

        This method takes into account that the Zope server and the browser 
        may be in different timezones.

        atime: a parsable date string
        """

    def latest():
        """
        Return the latest status update by the authenticated member
        """

class ICredentialsFormView(Interface):
    """
    Interface
    """

    def name():
        """
        Return name that browser view is registered with in ZCML
        """

    def submit(username, password, came_from=''):
        """
        Handle form submission
        """

class IPublishToTwitterFormView(Interface):
    """
    Interface
    """

    def name():
        """
        Return name that browser view is registered with in ZCML
        """

    def submit(came_from=''):
        """
        Handle form submission
        """

class ISylvesterFriendletManager(IViewletManager):
    """ 
    Render a friend
    """

class ISylvesterStatusletManager(IViewletManager):
    """ 
    Render a status
    """
